/** =================================================================\n
 *
 * Curve Compression Project - CS 7491 2004 project 1\n
 * @file Huffman.h
 *
 * Primary Authors: Nguyen Truong (nltruong@cc.gatech.edu)\n
 *                  Howard Zhou (howardz@cc.gatech.edu)\n
 *
 * (C) Copyright 2004\n
 * (C) Copyright Georgia Institute of Technology, 2004\n
 *
 * ==================================================================\n
 * Huffman compression
 *
 * @version 1.1\n
 *     
 * @date 2004-01-24
 * @author Nguyen Truong & Howard Zhou
 * 
 * @version 1.0\n
 *     Initial Revision
 * @date 2001-11-07
 * @author Howard Zhou
 *
 *
 * @todo
 * @bug
 * @warning 
 *
 * ===================================================================\n
 */

#ifndef __HUFFMAN_H__
#define __HUFFMAN_H__

#include <math.h>
#include <iostream>
#include <fstream>
#include <string>
#include <queue>

// local inclusion
#include "compress.h"
#include "mstring.h"
#include "node.h"

using namespace std;


// ==================================================================================
/**
 * @class CHuffman
 * @brief Huffman compression
 * @ingroup Compression
 *
 * @par requirements
 *
 * @par license
 * This code is absolutely free to use and modify. The code is provided "as is" with
 * no expressed or implied warranty. The author accepts no liability if it causes
 * any damage to your computer, causes your pet to fall ill, increases baldness
 * or makes your car start emitting strange noises when you start it up.
 * This code has no bugs, just undocumented features!
 *
 * 
 * @version 1.1\n
 *     
 * @date 2004-01-24
 * @author Nguyen Truong & Howard Zhou
 * 
 * @version 1.0\n
 *     Initial Revision
 * @date 2001-11-07
 * @author Howard Zhou
 *
 * This class is class that implements the huffman compression
 * 
 * @todo
 * @bug
 * @warning
 */
// ================================================================================

template <class T>
class CHuffman : public Compress
{
protected:
	// root of the Huffman binary tree
	CNode<T> *_root;  

private:
	// ------------------------------------------------------
	// private helpers
	// ------------------------------------------------------
	long encodeHuffmanTreeRecursive( CNode<T>* root, int bit, bit_file &output);
	void buildSymbolListRecursive(CNode<T> * node, vector<T> & symbols);
	double getEntropyRecursive(CNode<T>* node, double sum) const;

public:
	// ------------------------------------------------------
	// constructors and destructors
	// ------------------------------------------------------

	/// Default constructor. Creates an uninitialized matrix.
	CHuffman(){};

	//destructor
	~CHuffman(){};

	void CompressFile( String &input, String &output ){};
	void ExpandFile( String &input, String &output ){};

	// ------------------------------------------------------
	// Predicators
	// ------------------------------------------------------

	// ------------------------------------------------------
	// Accessors and Mutators
	// ------------------------------------------------------


	// ------------------------------------------------------
	// Overloaded Operators 
	// ------------------------------------------------------

	// ------------------------------------------------------
	// Other Operation 
	// ------------------------------------------------------
	void compressData(vector<T> data, T min, T max, bit_file&output);
	vector<T> decompressData(bit_file &input);
	
	void encodeSymbol(string strSymbol, bit_file& output);
	vector<T> decodeSymbol(CNode<T>* huffmanTree, vector<T> symbols, long nData, bit_file& input);

	long buildHuffmanTree( vector<T> data, T min, T max);
	long encodeHuffmanTree( bit_file &output);
	CNode<T>* decodeHuffmanTree( vector<T> symbols, long nBits, bit_file & input);

	vector<T> buildSymbolList();

	double getEntropy() const;

	// ------------------------------------------------------
	// Output or Test
	// ------------------------------------------------------

	// ------------------------------------------------------
	// Friend functions
	// ------------------------------------------------------

	//accessor and mutator
	CNode* getRoot()
	{	return _root; };

protected:
};

// ------------------------------------------------------
// Other Operation 
// ------------------------------------------------------
/**
 * decompressData
 *
 **/
template<>
vector<long> CHuffman<long>::decompressData(bit_file & in)
{
	vector<long> data;
	vector<long> symbols;
	CNode<long> *huffmanTree = NULL;

	long nData = -1;
	long nSymbols = -1;
	long treeSize = -1;

	long i;
	long symbol;	

	ifstream * inFile = in.iFile;

	*inFile >> nData;		// number of data points
	*inFile >> nSymbols;	// number of symbols
	*inFile >> treeSize;	// size of the encoded binary tree
	// cout << nData << endl << nSymbols << endl << treeSize;

	for(i = 0; i< nSymbols; i++)
	{
		*inFile >> symbol;
		// cout << symbol << endl;
		symbols.push_back(symbol);
	}

	huffmanTree = decodeHuffmanTree(symbols, treeSize, in);
	// huffmanTree -> display();

    data = decodeSymbol( huffmanTree, symbols, nData, in );
	return data;
}

/**
 * getEntropy
 *
 **/
template<class T>
double CHuffman<T>::getEntropy() const
{
	if(NULL == _root)
		return -1;

	double sum = _root->getWeightSum();
	// cout << sum << endl;

	return getEntropyRecursive(_root, sum);
}

/**
 * getEntropyRecursive
 *
 **/
template<class T>
double CHuffman<T>::getEntropyRecursive(CNode<T>* node, double sum) const
{
	double p, log2p;
	CNode<T> *left	= node->getLeft();
	CNode<T> *right = node->getRight();
	
	// at the internal node
	if(left && right)
	{
		return getEntropyRecursive(left, sum) + getEntropyRecursive(right, sum);
	}
	// at the leaf
	else
	{
		p = node->getWeight()/sum;
		log2p = log(p)/log(2);
		//cout << p << " -- " << log2p << " -- ";
		//cout << (- p * log2p) << endl;
		return (-p * log2p);
	}

}

/**
 * compressData
 *
 * @return double Entropy of 
 **/
template<>
void CHuffman<long>::compressData(vector<long> data, long min, long max, bit_file &out)
{
	// table to hold all possible occurance
	// assume the number of occurances fit in long type 
	long size = max - min +1;
	long treeSize = 0;
	long symbol;

	// first, we construct the huffman tree
	treeSize = buildHuffmanTree(data, min, max);

	string* symbolTable = new string[size];
	_root->buildSymbolTable(symbolTable, min);

	// ===================================================================	
	// now we try to compress the residuals
	ofstream * outFile = out.oFile;

	vector<long> symbols = buildSymbolList();


	*outFile << data.size() << endl;	// number of data points
	*outFile << symbols.size()<< endl;	// number of symbols
	*outFile << treeSize << endl;	// size of the encoded binary tree

	vector<long>::iterator curr;
	curr = symbols.begin();
	*outFile << *curr;
	curr++;
	for(curr;curr != symbols.end(); curr++)
	{
		*outFile << " " << *curr;
	}
	treeSize = encodeHuffmanTree(out);

	for(curr = data.begin(); curr != data.end(); curr++)
	{

		symbol = *curr;
		encodeSymbol(symbolTable[symbol-min], out);
	}

	_root->display();
	// cout << "\nEntropy: " << getEntropy() << endl;
}

/**
 * encodeHuffmanTree
 * a wrapper for the encodeHuffmanTreeRecursive which does the
 * real job of encoding the huffman tree recursively
 * 
 * @param output the output bit file
 * @return long the number of bits of the huffman tree
 *
 **/
template <class T>
long CHuffman<T>::encodeHuffmanTree( bit_file &output)
{
	return encodeHuffmanTreeRecursive(_root, 1, output);
}

/**
 * encodeHuffmanTreeRecursive
 * recursively encode the huffman tree from the give root node
 *
 * @param root the starting node
 * @param bit the bit content for the current node
 * @param output the output bit file
 * @return long the number of bits of the huffman tree
 *
 **/
template <class T>
long CHuffman<T>::encodeHuffmanTreeRecursive( CNode<T>* node, int bit, bit_file &output)
{
	CNode<T> *right;
	CNode<T> *left;

	OutputBit(output, bit);

	// taken out of if to avoid short circuit
	left = node->getRight();
	right = node->getLeft();

	// at internal nodes
	if (left && right)
	{
		return (1 + encodeHuffmanTreeRecursive(left, 0, output) + 
					encodeHuffmanTreeRecursive(right, 1, output));
	}
	// or else at the leaf
	return 1;
}

/**
 * buildSymbolList
 *
 **/
template <class T>
vector<T> CHuffman<T>::buildSymbolList()
{
	vector<T> symbols;
	buildSymbolListRecursive(_root, symbols);

	return symbols;
}

/**
 * buildSymbolListRecursive
 *
 **/
template <class T>
void CHuffman<T>::buildSymbolListRecursive(CNode<T> * node, vector<T> & symbols)
{
	CNode<T> *right;
	CNode<T> *left;


	// taken out of if to avoid short circuit
	left = node->getRight();
	right = node->getLeft();

	//  at the leaf
	if (NULL == left && NULL == right)
	{
		// cout << node->getSymbol() << endl;
		symbols.push_back(node->getSymbol());
	}
	// or else at internal nodes
	else
	{
		buildSymbolListRecursive(left, symbols);
		buildSymbolListRecursive(right, symbols);
	}
} 


/**
 * decodeHuffmanTree
 *
 **/
template <class T>
CNode<T>* CHuffman<T>::decodeHuffmanTree( vector<T> symbols, long nBits, bit_file &input )
{
	long i;
	int bit, nextBit;
	long symbolCount = 0;

	// root bit is 1
	bit = InputBit(input);
	CNode<T> * root = new CNode<T>(NULL, NULL, NULL, NULL, 0, "");
	CNode<T> * curr = root;

	for(i = 1; i < nBits; i++)
	{
		nextBit = InputBit(input);
		
		// 4 possible 2bits string: 00, 01, 10, 11
		if(0 == bit && 0 == nextBit)
		{
			// if the current bit is 0 and the next bit is also 0
			// then the current node is an internal node
			CNode<T> * n = new CNode<T>(NULL, NULL, NULL, curr, 0, curr->getCode() + "0");
			curr->setLeft(n);
			curr = n;
		}
		else if(0 == bit && 1 == nextBit)
		{
			// if the current bit is 0 and the next bit is 1
			// then the current node is a leaf
			curr->setSymbol(symbols[symbolCount]);
			symbolCount++;

			curr = curr->getParent();
			CNode<T> * n = new CNode<T>(NULL, NULL, NULL, curr, 0, curr->getCode() + "1");
			curr->setRight(n);
			curr = n;
		}
		else if(1 == bit && 0 == nextBit)
		{
			// if the current bit is 1 and the next bit is 0
			// the the current node is an internal node
			CNode<T> * n = new CNode<T>(NULL, NULL, NULL, curr, 0, curr->getCode() + "0");
			curr->setLeft(n);
			curr = n;
		}
		else // if(1 == bit && 1 == nextBit)
		{
			// if the current bit is 1 and the next bit is also 1
			// the the current node is a leaf
			// and we also need to rewind to the node which doesn't have a right child
			curr->setSymbol(symbols[symbolCount]);
			symbolCount++;

			curr = curr -> getParent();
			// rewind to the node which doesn't have a right child
			while(NULL != (curr ->getRight()))
				curr = curr->getParent();

			CNode<T> * n = new CNode<T>(NULL, NULL, NULL, curr, 0, curr->getCode() + "1");
			curr->setRight(n);
			curr = n;
		}

		bit = nextBit;
	}
	curr->setSymbol(symbols[symbolCount]);
	return root;
}


/**
 * buildHuffmanTree
 *
 * @param data the data points for compression
 * @param min, max propertie of these data points
 * @return long the number of bits needed to encode the huffman tree
 **/
template<>
long CHuffman<long>::buildHuffmanTree(vector<long>data, long min, long max)
{
	// the number of symbols that appeared at least once in data
	long nSymbols = 0;
	long treeSize = 0;
	long symbol;

	// construct the symbol first

	// table to hold all possible occurance
	// assume the number of occurances fit in long type 
	long size = max - min +1;
	long *symbolCountTable = new long[size];
	
	// initialize this array to '\0'
	memset(symbolCountTable, '\0', size*sizeof(long));

	vector<long>::iterator curr;
	for(curr = data.begin(); curr != data.end(); curr++)
	{
		symbol = *curr;

		// increase the count of symbol
		symbolCountTable[symbol-min]++;
	}

	// make a priority_queue and keep elements according to their weight
	priority_queue< CNode<long> > codeTable;


	CNode<long> *n1, *n2, *p;
	for(long i=0; i< size; i++)
	{
		// count the number of char that appeared at least once
		if (symbolCountTable[i])
		{
			symbol = i + min;
			nSymbols++;

			// build a node and insert this node into the priority_queue
			codeTable.push(CNode<long>(symbol, NULL, NULL, NULL, symbolCountTable[i], ""));
			treeSize ++;
		}
	} 

	//test the contents of the priority_queue
	while(!codeTable.empty())
	{
		n1 = new CNode<long>();
		n2 = new CNode<long>();

		// the least possible argument
		*n1 = codeTable.top();
		codeTable.pop();

		if (!codeTable.empty())
		{  
			// the second least possible argument
			*n2 = codeTable.top();
			codeTable.pop();

			p = new CNode<long>(size, n2, n1, NULL, n1->getWeight() + n2->getWeight(), "");
			// push the merged node into the priority_queue
			codeTable.push(*p);
			n1->setParent(p);
			n2->setParent(p);
			treeSize++;
		}
		else
		{
			// emptied priority queue
			_root = n1;
		}
	}

	// recursively assign Huffman code
	_root->setCodeAll();

	// recursively output the code table to a file
	// _root->display();

	cout << "\nnumber of Symbols appeared at least once: " << nSymbols << endl;
	//cout << "\nSource file size: " << srcFileSize;
	//cout << "\nCompressed file size: " << (encodedFileSize = _root -> getCompressedSize(ASCIITable)/8);
	//cout << "\nCode table size: " << (codeTableSize = _root -> getCodeTableSize(ASCIITable))/8;
	//cout << "\nCompression ratio: " << ((double)(encodedFileSize + codeTableSize)/(double)srcFileSize)*100.0 << endl;

	return treeSize;

}




/**
 * encodeSymbol
 *
 **/
template <class T>
void CHuffman<T>::encodeSymbol( string strSymbol, bit_file &output )
{
	// cout << strSymbol << endl;
	char c;

	string::iterator curr;
	for(curr = strSymbol.begin(); curr != strSymbol.end(); curr++)
	{
		c = char(*curr);
		// cout << c << '-';

		if (c == '0')
			OutputBit(output, 0);
		else
			OutputBit(output, 1);
	}
}


/**
 * decodeSymbol
 * following the binary try
 * 
 **/
template <>
vector<long> CHuffman<long>::decodeSymbol(CNode<long>* tree, vector<long> symbols, long nData, bit_file &input)
{
	vector<long> data;
	long symbol;
	int bit;
	long count = 0;
	CNode<long> * node;

	while (!input.IsEOF() && count < nData)
	{
		node = tree;
		while(! node->isLeaf())
		{
			bit = InputBit(input);

			if(0 == bit)
			{
				node = node->getLeft();
			}
			else
			{
				node = node->getRight();
			}
		}
		symbol = node->getSymbol();
		data.push_back(symbol);
		// cout << symbol << endl;
		count ++;
	}
	return data;
}



#endif	// __HUFFMAN_H__
