// mstring.h

/*
	replacement string class for cstring.h defined by
	Borland. Very straight-forward char array based
	string implementation. Leaves out some of the fancier
	functions for a tighter class.
*/

#ifndef __cplusplus
	#error Must use C++ for MSTRING.H
#endif

//thread safety includes...
#ifndef _WINDOWS_
	#ifdef WIN32
		#include <windows.h>
	#endif
#endif

//define sentinal
#ifndef _MSTRING_H
	#define _MSTRING_H
//bring in dependancy headers
#ifdef _MSC_VER
    #ifndef _INC_CTYPE
    	#include <ctype>
    #endif
    #ifndef _INC_IOSTREAM
    	#include <iostream>
    #endif
    #ifndef _INC_STRING
    	#include <string>
    #endif
    #ifndef _INC_STDLIB
    	#include <stdlib>
    #endif
    #ifndef _INC_NEW
    	#include <new>
    #endif
#else
    #ifndef __EXCEPT_H
	    #include <except>
    #endif
    #ifndef __CTYPE_H
    	#include <ctype>
    #endif
    #ifndef __IOSTREAM_H
    	#include <iostream>
    #endif
    #ifndef __STRING_H
    	#include <string>
    #endif
    #ifndef __STDLIB_H
    	#include <stdlib>
    #endif
    #ifndef __NEW_H
    	#include <new>
    #endif
#endif
#ifndef	BOOL
	#define BOOL	short
#endif


//maximum length for temp strings
#define 	MAX_TMP_LEN	1024

using namespace std;

//MACRO for ptr creation & validation
#define	NEW_PTR(p,l)	p = new char[l + 1];  \
						if(p == NULL) {       \
							throw ("String::memory allocation failed"); } \
						memset(p,0,l+1);
//MACRO for safe removal of old ptr
#define	DEL_PTR(p)	if(p != NULL)   {\
                        delete [] p; \
	    				p = NULL;   }

/*           
    the actual string class itself
*/
class String {
	/*
    the simple string node object, (inline class),
    contains some basic operators to 
    simplify the code later, it is never really externalized
	I may modifiy this later to simply use a PASCAL string format.
	attempts to mutex access to the string data, (critical section
	on WIN32)
	*/
	class str_node	{
	public:
		  str_node ()
	      {
		      _str = NULL;
			  _len = 0;
#ifdef	WIN32
			  InitializeCriticalSection(&m_CS);	
#endif
	      }

		  str_node (str_node& s)
	      { 
			  _str = NULL;
			  _len = 0;
		      if(s._str != NULL)    {
				if(s._str[0] != 0x00)	{
					_len = s._len;
					NEW_PTR(_str,s._len)
			        strcpy(_str,s._str);
				 }
		      }
	      }
		virtual ~ str_node () { //be certain we have cleaned up 
					GrabMutex();
					DEL_PTR(_str)
		   	         _len = 0;
					FreeMutex();
#ifdef WIN32
					DeleteCriticalSection(&m_CS);
#endif
				}

	    int isnull(void) const
	    {
			//bum pointer
			if(_str == NULL)	{
				return  1;
			}
			//address valid, but still bad?
			if(_str[0] == 0x00)	{
				return 1;
			}
			return 0;
	    }

	    void clear(void)
	    {
			GrabMutex();
	         if(!isnull())  {
	             DEL_PTR(_str)
	             _len = 0;
	         }
	         else   {
	             _str = NULL;
	             _len = 0;
	         }
			 FreeMutex();
	    }

	    str_node &operator = (const str_node &y)
	    {
			  GrabMutex();
	          if(!isnull())  {
	              clear();
	          }
	          if(!y.isnull())    {
	              _len = y._len;
	              NEW_PTR(_str,_len - 1)
	              strcpy(_str,y._str);
	          }
			  FreeMutex();
          return (*this);
	    }

	    str_node &operator = (const char *y)
	    {
			 GrabMutex();
	          if(!isnull())  {
	              clear();
	          }
	          if((y != NULL) && (y[0] != 0x00))   {
	              _len = strlen(y) + 1;
	              NEW_PTR(_str,strlen(y))
	              strcpy(_str,y);
	          }
			  FreeMutex();
	          return (*this);
	    }
	
	    operator char *()
	    {
	        return _str;
	    }
	
	    operator unsigned char *()
	    {
	        return (unsigned char *)_str;
	    }
	
		char operator [](int index)
	    {
	        char nil = '\0';
	
	        if((isnull()) || (index > _len) || (index < 0))
	            return nil;
	        else
	            return _str[index];
	    }
	
		void GrabMutex(void)
		{
#ifdef WIN32
			EnterCriticalSection(&m_CS);
#endif
		}

		void FreeMutex(void)
		{
#ifdef WIN32
			LeaveCriticalSection(&m_CS);
#endif
		}

		//variables...
		char *_str;
		int  _len;
#ifdef WIN32
		CRITICAL_SECTION	m_CS;
#endif
	};
	/*	   end of embedded class	*/


protected:
	    char     nilchar;   //NULL character
		str_node str;       //the string object
		int		 m_case;	//case sensitivity flag
public:
		//default constructor
		  String ();
		//one arg
		  String (const char  *instring);
 virtual ~ String ();
		// copy constructor(s)
		String (const String &other);
		String (const String &x, int s, int c) ;
		/*public vars*/
		enum	StripType { Leading, Trailing, Both };
		/*    operators next    */
		//pretend to be a char *
				operator char  *();
				operator unsigned char *();
		// overload the assignment operator
		String  &operator = (const String &other);
		String  &operator = (const char *y);
		String  &operator += (const String &other);
		String  &operator += (const char *other);
		String  &operator -= (const String &other);
		String  &operator -= (const char *other);
        //math operators...
//		String   operator + (const String &aa, const String &bb);
//		String   operator + (const char  *aa, const String &bb);
//		String   operator + (const String &aa, const char *bb);
//		String   operator + (const char  *aa, const char *bb);
//		String   operator - (const String &aa, const String &bb);
//		String   operator - (const char  *aa, const String &bb);
//		String   operator - (const String &aa, const char *bb);
//		String   operator - (const char  *aa, const char *bb);
        //streaming
friend ostream  &operator<< (ostream& os, String& aString)
{
  if(aString.c_str() != NULL)
      os << aString.c_str();
  return os;
}

friend istream& operator>> (istream& is, String& aString)
{
	//up to MAX_TMP_LEN strings may be streamed in
	char *	buf;
                              
	NEW_PTR(buf,MAX_TMP_LEN)
	is >> buf;
	aString.str = buf;
	DEL_PTR(buf)
	return is;
}

        //comparison
		BOOL 	operator == (String &ss);
		BOOL 	operator == (const char  *ss);
		BOOL 	operator != (String &ss);
		BOOL 	operator != (const char  *ss);
		BOOL 	operator > (String ss);
		BOOL 	operator > (const char * ss);
		BOOL 	operator < (String ss);
		BOOL 	operator < (const char * ss);
		BOOL 	operator <= (String ss);
		BOOL 	operator <= (const char * ss);
		BOOL 	operator >= (String ss);
		BOOL 	operator >= (const char * ss);
        //index operators
  	    char 	operator [](int index);
        char 	operator ()(int index);
		String	operator () (int index, int n);
        //now some bitwise operators...
		String  &operator |= (String &other);
		String  &operator &= (String &other);
		String  &operator ^= (String &other);

		friend ostream& operator<< (ostream& os, String& aString);
		friend istream& operator>> (istream& is, String& aString);



		/*basic public functions */
		int  	allocate( int size, const char fc = '\0' );
        int 	append( const String &s );
		int 	append( const char  ss );
		int 	append( const String &s, int start, int n = -1 );
		int 	append( const char  *cp, int start, int n = -1 );
		int		append( const char  *ss );
		int		assign( String &s );
		int		assign( const char  *s );
		int 	assign( String &s, int start, int n = -1 );
		int 	assign( const char  *s, int start, int n = -1 );
		void	case_toggle(void);
		int	 	chrcnt(const char s);
		void 	clear(void)	{ str.clear(); }
		void 	clone(String *ss);
		void 	clone(String &ss);
		int 	compare(String &s);
		int 	compare(String &s, int orig, int n = -1 );
		int 	compare(const char  * s);
		int 	compare(char  * s, int orig, int n = -1 );
		int 	contains(const char  * pat);
		int 	contains(String & s);
		int 	copy( char  *cb, int n = -1 );
		int 	copy( char  *cb, int n, int pos );
		//note the following will cheerfully return NULL
		char   *c_str() const	{ return str._str; }
		void 	empty(void)	{ str.clear(); }
		int 	find( String &s );
		int 	find( String &s, int pos );
		int 	find( const char  *s );
		int 	find( const char  *s, int pos );
		int		find_first_of( String &s );
		int		find_first_of( String &s, int pos );
		int		find_first_of( const char *s );
		int 	find_first_of( const char *s, int pos );
		int 	find_first_not_of( String &s );
		int 	find_first_not_of( String &s, int pos );
		int 	find_first_not_of( const char *s );
		int 	find_first_not_of( const char *s, int pos );
		int 	find_last_of( String &s );
		int 	find_last_of( String &s, int pos );
		int 	find_last_of( const char *s );
		int 	find_last_of( const char *s, int pos );
		int 	find_last_not_of( String &s );
		int 	find_last_not_of( String &s, int pos );
		int 	find_last_not_of( const char *s );
		int 	find_last_not_of( const char *s, int pos );
		char 	get_at(int pos);
		int	 	getlen(void) const { return ((str._len == 0) ? str._len : (str._len - 1)); }
		void	getline(istream &stream);
unsigned long	hash(void);
		int 	insert( int pos, String &s );
		int 	insert( int pos, String &s, int start, int n = -1 );
		int 	insert( int pos, const char  *s );
		int 	insert( int pos, char  *s, int start, int n = -1 );
		int  	is_null(void) { return str.isnull(); }
		String	left(int len); // return first 'len' characters
		int 	length() { return getlen(); }
		String	mid(int start, int len); // return 'len' characters from offset 'start'
		int 	prepend( String &s );
		int		prepend( String &s, int start, int n = -1 );
		int		prepend( const char  *cp );
		int		prepend( const char  *cp, int start, int n = -1 );
		int  	put_at( int pos, char c );
		istream& read_string(istream &stream);
//		istream& read_file(istream &stream);
//		istream& read_to_delim(istream &stream, char delim = '\n');
//		istream& read_token(istream &stream);
		istream& read_line(istream &stream);
		int		remove( int pos, int n = -1 );
		void	remove_all_of(String &set);
		void	remove_all_of(const char *set);
		void	remove_all_not_of(String &set);
		void	remove_all_not_of(const char *set);
		int 	replace( int pos, String &s, int n = -1 );
		int 	replace( int pos, int n1, String &s, int start, int n2 );
		int 	replace( int pos, const char * s, int n = -1 );
		int 	replace( int pos, int n1, const char * s, int start, int n2 );
		int	 	reserve(void);
		void 	resize(int m);
		void 	reverse(void);
		int 	rfind( String &s );
		int 	rfind( String &s, int pos );
		int 	rfind( const char * s );
		int 	rfind( const char * s, int pos );
		String	right(int len); // return last 'len' characters
		int		search(const char  *word);
		void 	set_at(int pos, char c);
		void	set_case_sensitive(int tf = 1)	{ m_case = tf; }
		void 	strrpl(const char o, const char n);
		void 	strset(const char s);
		int		strindex(const char  *word);
		void	strip(StripType t = Trailing);
		int		strstr_cnt(const char *s);
		int		strstr_cnt(String &s);
		int		substrrpl(String &SubOld, String &SubNew);
		int		substrrpl(String &SubOld, const char *SubNew);
		int		substrrpl(const char *SubOld, String &SubNew);
		int		substrrpl(const char *SubOld, const char *SubNew, int num_repl = 0);
		String	substr( int pos );
		String	substr( int pos, int n = -1 );
		void 	to_lower(void);
		void 	to_upper(void);
protected:
		String	add(const char *aa, const char*bb);
		void	GrabMutex(void);
		void	FreeMutex(void);
		int 	removechars( int pos );
		int		search(const char  *searchstr, const char  *word);
		void 	striplead(void);
		void	striptail(void);
};

#endif
