/*
 * This is the LZ77 class, which implements an LZ77 style compression
 * algorithm.  As iplemented here it uses a 12 bit index into the sliding
 * window, and a 4 bit length, which is adjusted to reflect phrase lengths
 * of between 2 and 17 bytes.
*/
#ifndef _COMPRESSLZ77_H
	#define _COMPRESSLZ77_H
#ifndef _COMPRESS_H
	#include "compress.h"
#endif
#ifndef _MSTRING_H
	#include "mstring.h"
#endif
#ifndef _INC_STDIO
	#include <stdio.h>
#endif
#ifndef __IOSTREAM_H
	#include <iostream>
#endif

using namespace std;

//some global definitions...
/*
 * Various constants used to define the compression parameters.  The
 * INDEX_BIT_COUNT tells how many bits we allocate to indices into the
 * text window.  This directly determines the WINDOW_SIZE.  The
 * LENGTH_BIT_COUNT tells how many bits we allocate for the length of
 * an encode phrase. This determines the size of the look ahead buffer.
 * The TREE_ROOT is a special node in the tree that always points to
 * the root node of the binary phrase tree.  END_OF_STREAM is a special
 * index used to flag the fact that the file has been completely
 * encoded, and there is no more data.  UNUSED is the null index for
 * the tree. MOD_WINDOW() is a macro used to perform arithmetic on tree
 * indices.
 *
 */

#define INDEX_BIT_COUNT      12
#define LENGTH_BIT_COUNT     4
#define WINDOW_SIZE          ( 1 << INDEX_BIT_COUNT )
#define RAW_LOOK_AHEAD_SIZE  ( 1 << LENGTH_BIT_COUNT )
#define BREAK_EVEN           ( ( 1 + INDEX_BIT_COUNT + LENGTH_BIT_COUNT ) / 9 )
#define LOOK_AHEAD_SIZE      ( RAW_LOOK_AHEAD_SIZE + BREAK_EVEN )
#define TREE_ROOT            WINDOW_SIZE
#define END_STREAM		     0
#define UN_USED              0
#define MOD_WINDOW( a )      ( ( a ) & ( WINDOW_SIZE - 1 ) )

class CompressLZ77	: public Compress {
public:
	  CompressLZ77 ();
	~ CompressLZ77 ();
	void CompressFile( String &input, String &output );
	void ExpandFile( String &input, String &output );

protected:

	void InitTree( int r );
	void ContractNode( int old_node, int new_node );
	void ReplaceNode( int old_node, int new_node );
	int  FindNextNode( int node );
	void DeleteString( int p );
	int  AddString( int new_node, int *match_position );
/*
 * These are the two global data structures used in this program.
 * The window[] array is exactly that, the window of previously seen
 * text, as well as the current look ahead text.  The tree[] structure
 * contains the binary tree of all of the strings in the window sorted
 * in order.
*/
    unsigned char window[ WINDOW_SIZE ];

    struct ltree {
        int parent;
        int smaller_child;
        int larger_child;
    } tree[ WINDOW_SIZE + 1 ];
};
#endif