#include <stdio.h>
#include "mstring.h"
#include "bitfile.h"
#include "compress.h"
#ifdef	WIN32
	#ifndef _CONSOLE
		#include <windows.h>
	#endif
#endif


bit_file *Compress::OpenOutputBitFile( String &name )
{
	bit_file *Out;

    Out = new bit_file(bit_file::Output, name);

    Out->rack = 0;
    Out->mask = 0x80;
    Out->pacifier_counter = 0;
    return( Out );
}

bit_file *Compress::OpenInputBitFile( String &name )
{
    bit_file *In;

    In = new bit_file(bit_file::Input, name);

    In->rack = 0;
    In->mask = 0x80;
    In->pacifier_counter = 0;
    return( In );
}

void Compress::CloseOutputBitFile( bit_file *m )
{
    if ( m->mask != 0x80 )	{
        m->CloseFile();
	}
    delete m;
}

void Compress::CloseInputBitFile( bit_file *m )
{
    m->CloseFile();
    delete m;
}

void Compress::OutputBit( bit_file &m, int bit )
{
    if ( bit )
        m.rack |= m.mask;
    m.mask >>= 1;
    if ( m.mask == 0 ) {
		m.PutBit(m.rack);

		m.rack = 0;
		m.mask = 0x80;
		if ( ( m.pacifier_counter++ & PACIFIER_COUNT ) == 0 )	{
#ifndef	WINVER	//not a WIN16 or WIN32 app
			putc( typeChar, stdout );
#else
	#ifndef _CONSOLE	//not a WIN32 Console app
			MessageBeep(MB_OK);
	#else
			putc( typeChar, stdout );
	#endif
#endif
		}
    }
}

void Compress::OutputBits( bit_file &m, unsigned long code, int count )
{
    unsigned long   mask;

    mask = 1L << ( count - 1 );
    while ( mask != 0) {
        if ( mask & code )
            m.rack |= m.mask;
        m.mask >>= 1;

        if ( m.mask == 0 ) {
            m.PutBit(m.rack);
				if ( ( m.pacifier_counter++ & PACIFIER_COUNT ) == 0 )	{
#ifndef	WINVER	//not a WIN16 or WIN32 app
					putc( typeChar, stdout );
#else
	#ifndef _CONSOLE	//not a WIN32 Console app
					MessageBeep(MB_OK);
	#else
					putc( typeChar, stdout );
	#endif
#endif
			}
			m.rack = 0;
			m.mask = 0x80;
		}
        mask >>= 1;
    }
}

int Compress::InputBit( bit_file &m )
{
    int value;

    if ( m.mask == 0x80 ) {
        m.rack = m.GetBit();
        if ( m.IsEOF() )
            throw( "Fatal error in InputBit!\n" );
    if ( ( m.pacifier_counter++ & PACIFIER_COUNT ) == 0 )
#ifndef	WINVER	//not a WIN16 or WIN32 app
			putc( typeChar, stdout );
#else
	#ifndef _CONSOLE	//not a WIN32 Console app
			MessageBeep(MB_OK);
	#else
			putc( typeChar, stdout );
	#endif
#endif
    }
    value = m.rack & m.mask;
    m.mask >>= 1;
    if ( m.mask == 0 )
		m.mask = 0x80;
    return( value ? 1 : 0 );
}

unsigned long Compress::InputBits( bit_file &m, int bit_count )
{
    unsigned long mask;
    unsigned long return_value;

    mask = 1L << ( bit_count - 1 );
    return_value = 0;

    while ( mask != 0) {
	if ( m.mask == 0x80 ) {
	    m.rack = m.GetBit();
	    if ( m.IsEOF() )
			throw( "Fatal error in InputBit!\n" );
        if ( ( m.pacifier_counter++ & PACIFIER_COUNT ) == 0 )
#ifndef	WINVER	//not a WIN16 or WIN32 app
			putc( typeChar, stdout );
#else
	#ifndef _CONSOLE	//not a WIN32 Console app
			MessageBeep(MB_OK);
	#else
			putc( typeChar, stdout );
	#endif
#endif
	}
	if ( m.rack & m.mask )   {
            return_value |= mask;
    }
        mask >>= 1;
        m.mask >>= 1;
        if ( m.mask == 0 )
            m.mask = 0x80;
    }
    return( return_value );
}

int Compress::OutputStd( int bit, bit_file *m )
{
    m->PutBit(bit);
    return bit;
}

int Compress::InputStd( bit_file *m )
{
    return m->GetBit();
}

