/** =================================================================\n
 *
 * Curve Compression Project - CS 7491 2004 project 1\n
 * @file Huffman.h
 *
 * Primary Authors: Nguyen Truong (nltruong@cc.gatech.edu)\n
 *                  Howard Zhou (howardz@cc.gatech.edu)\n
 *
 * (C) Copyright 2004\n
 * (C) Copyright Georgia Institute of Technology, 2004\n
 *
 * ==================================================================\n
 * Huffman compression
 *
 * @version 1.1\n
 *     
 * @date 2004-01-24
 * @author Nguyen Truong & Howard Zhou
 * 
 * @version 1.0\n
 *     Initial Revision
 * @date 2001-11-07
 * @author Howard Zhou
 *
 *
 * @todo
 * @bug
 * @warning 
 *
 * ===================================================================\n
 */

#ifndef __NODE_H__
#define __NODE_H__

#include <string>
#include <iostream>


using namespace std;

// ==================================================================================
/**
 * @class CHuffman
 * @brief Huffman compression
 * @ingroup Compression
 *
 * @par requirements
 *
 * @par license
 * This code is absolutely free to use and modify. The code is provided "as is" with
 * no expressed or implied warranty. The author accepts no liability if it causes
 * any damage to your computer, causes your pet to fall ill, increases baldness
 * or makes your car start emitting strange noises when you start it up.
 * This code has no bugs, just undocumented features!
 *
 * 
 * @version 1.1\n
 *     
 * @date 2004-01-24
 * @author Nguyen Truong & Howard Zhou
 * 
 * @version 1.0\n
 *     Initial Revision
 * @date 2001-11-07
 * @author Howard Zhou
 *
 * This class is class that implements the huffman compression
 * 
 * @todo
 * @bug
 * @warning
 */
// ================================================================================

//--------------------------------------------------------------------------
// notes:
//    each node will have one left child and one right child (of node type) 
//    and contains a character and a number (weight) indicating the number of 
//    occurance that character in a specific file
//
//--------------------------------------------------------------------------

template<class T>
class CNode
{
private:
	T _symbol;

	CNode *_right;
	CNode *_left;
	CNode *_parent;
	long _weight;
	string _code;

public:

	// ------------------------------------------------------
	// constructors and destructors
	// ------------------------------------------------------
	
	// override default constructor
	CNode(T symbol=0,CNode *right=NULL, CNode *left=NULL, CNode *parent=NULL,
		 long weight=0, string code=""):
		_symbol	(symbol),
		_right	(right),
		_left	(left),
		_parent (parent),
		_weight	(weight),
		_code	(code)
	{};

	// copy constructor
	CNode(const CNode &node):
		_symbol	(node.getSymbol()),
		_right	(node.getRight()),
		_left	(node.getLeft()),
		_parent (node.getParent()),
		_weight	(node.getWeight()),
		_code	(node.getCode())
	{};


	//destructor
	~CNode(){};

	// ------------------------------------------------------
	// Predicators
	// ------------------------------------------------------
	bool isLeaf()
	{
		return( NULL == getRight() && NULL == getLeft());
	}

	// ------------------------------------------------------
	// Accessors and Mutators
	// ------------------------------------------------------
	CNode *getRight	() const {return _right; };
	CNode *getLeft	() const {return _left;  };
	CNode *getParent() const {return _parent;};
	T getSymbol		() const {return _symbol;};
	long getWeight	() const {return _weight;};
	string getCode	() const {return _code;  };
	double getWeightSum () const;
  
	void setRight	(CNode *n)		{_right	= n;};
	void setLeft	(CNode *n)		{_left	= n;};
	void setParent	(CNode *p)		{_parent= p;};
	void setSymbol	(T symbol)		{_symbol= symbol;};
	void setWeight	(long weight)	{_weight= weight;};
	void setCode	(string code)	{_code	= code;};


	// ------------------------------------------------------
	// Overloaded Operators 
	// ------------------------------------------------------
	bool operator< (const CNode<T>& n) const
	{
		// we want lowest weight CNode on top of the priority_queue
		// since we use this function to replace the default less function
		return (_weight > n.getWeight());
	}

	// ------------------------------------------------------
	// Other Operation 
	// ------------------------------------------------------
	// recursively build the symbol table
	void buildSymbolTable(string * table, long min)
	{
		CNode *right;
		CNode *left;

		right = getRight();
		left = getLeft();

		// at internal nodes
		if (left && right)
		{
			left->buildSymbolTable(table, min);
			right->buildSymbolTable(table, min);
			return;
		}
		// at the leaf, store the symbol to the symbol table
		
		table[this->_symbol - min] = this->_code;
		return;
	}

	// recursively set the huffman's code
	void setCodeAll()
	{
		CNode *right = getRight();
		CNode *left = getLeft();

		// at internal nodes
		if (left && right)
		{
			left->setCode(getCode() + "1");
			right->setCode(getCode() + "0");
			left->setCodeAll();
			right->setCodeAll();
		}
		// at the leaf, do nothing
	}   
  
	// recursively get the sum of all weights
	double getWeightSum()
	{
		CNode *right = getRight();
		CNode *left = getLeft();

		if (left && right)
		{
			return (left->getWeightSum() + right->getWeightSum());
		}
		// at the leaf
		else
		{
			return getWeight();
		}
	}


	// ------------------------------------------------------
	// Output or Test
	// ------------------------------------------------------
	// recursively display the whole tree structure if there is any
	void display ()
	{
		CNode *right;
		CNode *left;

		right= getRight();
		left = getLeft();
		// at internal nodes
		if (left && right)
		{
			left->display();
			right->display();
		}
		// at the leaf
		else 
		{
			cout << *this;
		} 
	};


	// ------------------------------------------------------
	// Friend functions
	// ------------------------------------------------------
	friend ostream& operator<<(ostream & os, const CNode & node) 
	{
		return os << "char: " << node.getSymbol()
			  << " | code: " << node.getCode()
			  << " | weight: " << node.getWeight() << endl;
	}


	//friend class to compare node in priority_queue
	//this class must have a () operator and a empty constructor
	//friend class NodeCompare;
    
};


/*
//friend class to compare node in priority_queue
//this class must have a () operator and a empty constructor
class NodeCompare
{
public:
	//empty constructor (not overwritten)
	//() operator
	bool operator() (const CNode& x, const CNode& y) const 
	{
		// we want lowest weight CNode on top of the priority_queue
		// since we use this function to replace the default less function
		return(x._weight > y._weight);
	};
};
*/

#endif	// __NODE_H__
