/** =================================================================\n
 *
 * Curve Compression Project - CS 7491 2004 project 1\n
 * @file test.cpp
 *
 * Primary Authors: Nguyen Truong (nltruong@cc.gatech.edu)\n
 *                  Howard Zhou (howardz@cc.gatech.edu)\n
 *
 * (C) Copyright 2004\n
 * (C) Copyright Georgia Institute of Technology, 2004\n
 *
 * ==================================================================\n
 * Test case for polygonal curve compression and simplification
 *
 * @version 1.0\n
 *     Initial Revision
 * @date 2004-02-03
 * @author Nguyen Truong & Howard Zhou
 *
 *
 * @todo
 * @bug
 * @warning 
 *
 * ===================================================================\n
 */

#include <stdio.h>
#include <iostream>

#include "Compression\comphuff.h"
#include "Curve\PolygonalCurve.h"

#define HIGHEST_PRECISION 10

using namespace std;

void test1(void);
void test2(void);
void test3(void);

int main(int argc, char * argv[])
{
	test1();
	// test2();
	// test3();
	return 0;
}


void test3(void)
{
	CPolygonalCurve<CPoint2DReal> curve;
	CPolygonalCurve<CPoint2DReal> newCurve;


	String infile = "data\\c1p\\c_compressed.txt";
	String outfile = "data\\c1p\\c_cc.txt";
	String outfile2 = "data\\c1p\\c_out.txt";

	CompressHuff huff;
	huff.CompressFile(infile, outfile);

	CompressHuff huff1;
	huff1.ExpandFile(outfile, outfile2);
}


void test1(void)
{
	printf("Curve Compression\n");

	string filename;
	int nXBits, nYBits;
	long predictionFlag;
	int nSubdivision;
	int i;

	// 0: bSpline, 1: 4-point, 2: jarek
	int subdivisionMethod;

	ifstream script;

	script.open("data\\CurveList.txt",ios::in);
	if (script.bad())
		exit(-1);

	while(!script.eof())
	{
		CPolygonalCurve<CPoint2DReal> initCurve;
		CPolygonalCurve<CPoint2DReal> curve;
		CPolygonalCurve<CPoint2DReal> newCurve;
		CPolygonalCurve<CPoint2DReal> ctemp;
		CPolygonalCurve<CPoint2DReal> ctemp1;

		script >> filename >> nSubdivision >> subdivisionMethod;
		if (filename.size() == 0)
			break;

		cout << "\n====================================================\n";
		cout << "Processing file :" << filename << ".txt" << endl;

		string inputFilename = filename + ".txt";
		string inputFilenameDraw = filename + "_draw.txt";
		string origFilename = filename + "_orig.txt";
		string origFilenameDraw = filename + "_orig_draw.txt";
		string compressedFilename = filename + "_compressed.txt";
		string newFilename = filename + "_.txt";
		string newFilenameDraw = filename + "_new_draw.txt"; 

		initCurve.readFile(inputFilename.c_str());
		initCurve.draw(inputFilenameDraw.c_str());

		// construct the originalcurve
		ctemp = initCurve;
		for(i = 0; i < nSubdivision; i++)
		{
			ctemp1 = ctemp.subdivided(enum SubdivisionMethod(subdivisionMethod));
			ctemp = ctemp1;
		}

		curve =ctemp;
		curve.writeFile(origFilename.c_str());
		curve.draw(origFilenameDraw.c_str());

		// input the simplification bits and scheme
		script >> nXBits >> nYBits >> predictionFlag;

		// compressed the simplified curve
		curve.writeCompressedFile(compressedFilename.c_str(), CPoint2DReal(nXBits, nYBits), predictionFlag);
		curve.displayStatistics();

		/*
		char buf[255];
		strcpy(buf, filename.c_str());
		strcat(buf, "_compressed.txt");
		String infile(buf);

		strcpy(buf, filename.c_str());
		strcat(buf, "_cc.txt");
		String outfile(buf);

		strcpy(buf, filename.c_str());
		strcat(buf, "_compressed1.txt");
		String outfile2(buf);

		CompressHuff huff;
		huff.CompressFile(infile, outfile);

		CompressHuff huff1;
		huff1.ExpandFile(outfile, outfile2);
		*/

		newCurve.readCompressedFile(compressedFilename.c_str());
		newCurve.writeFile(newFilename.c_str());
		newCurve.draw(newFilenameDraw.c_str());

		cout << "\n====================================================\n";
	}
	script.close();
}

void test2(void)
{
	printf("Curve Simplification and Error\n");
	
	string filename;
	int nXBits, nYBits;
	long predictionFlag;
	int nSubdivision;
	int i;
	char bufferX[20];
	char bufferY[20];

	// 0: bSpline, 1: 4-point, 2: jarek
	int subdivisionMethod;

	ifstream script;

	script.open("data\\CurveList2.txt",ios::in);
	if (script.bad())
		exit(-1);

	while(!script.eof())
	{
		CPolygonalCurve<CPoint2DReal> initCurve;
		CPolygonalCurve<CPoint2DReal> curve;
		CPolygonalCurve<CPoint2DReal> newCurve;
		CPolygonalCurve<CPoint2DReal> bSpline;
		CPolygonalCurve<CPoint2DReal> fourPoint;
		CPolygonalCurve<CPoint2DReal> jarek;
		CPolygonalCurve<CPoint2DReal> simplified;
		CPolygonalCurve<CPoint2DReal> simplifiedNew;
		CPolygonalCurve<CPoint2DReal> ctemp;
		CPolygonalCurve<CPoint2DReal> ctemp1;

		script >> filename >> nSubdivision >> subdivisionMethod;
		if (filename.size() == 0)
			break;

		cout << "\n====================================================\n";
		cout << "Processing file :" << filename << ".txt" << endl;

		string inputFilename = filename + ".txt";
		string inputFilenameDraw = filename + "_draw.txt";
		string origFilename = filename + "_orig.txt";
		string origFilenameDraw = filename + "_orig_draw.txt";

		initCurve.readFile(inputFilename.c_str());
		initCurve.draw(inputFilenameDraw.c_str());

		// construct the originalcurve
		ctemp = initCurve;
		for(i = 0; i < nSubdivision; i++)
		{
			ctemp1 = ctemp.subdivided(enum SubdivisionMethod(subdivisionMethod));
			ctemp = ctemp1;
		}

		curve =ctemp;
		curve.writeFile(origFilename.c_str());
		curve.draw(origFilenameDraw.c_str());


		// input the simplification bits and scheme
		script >> nXBits >> nYBits >> predictionFlag;

		cout << nXBits << " simplification ";
		cout << "\n-------------------------------------\n";

		// simplify the curve using B-bit quantization and removing all 
		// vertices that are identical to their previous neighbor
		CPoint2DReal nBits(nXBits, nYBits);
		simplified = curve.simplified(nBits);


		itoa(nXBits, bufferX, 10);
		itoa(nYBits, bufferY, 10);
		// output the simplified curve
		string simpFilename = filename + "_sim_" + string(bufferX) + "_" + string(bufferY);
		string simpFilenameOut = simpFilename + ".txt";
		string simpFilenameDraw = simpFilename + "_draw.txt";
		string bSplineFilename = simpFilename + "_b.txt";
		string bSplineFilenameDraw = simpFilename + "_b_draw.txt";
		string fourPointFilename = simpFilename + "_4.txt";
		string fourPointFilenameDraw = simpFilename + "_4_draw.txt";
		string jarekFilename = simpFilename + "_j.txt";
		string jarekFilenameDraw = simpFilename + "_j_draw.txt";
		string simpFilenameCompressed = simpFilename + "_compressed.txt";


		simplified.writeFile(simpFilenameOut.c_str());
		simplified.draw(simpFilenameDraw.c_str());

		// compressed the simplified curve
		simplified.writeCompressedFile(simpFilenameCompressed.c_str(), CPoint2DReal(10, 10), predictionFlag);
		simplified.displayStatistics();

		simplifiedNew.readCompressedFile(simpFilenameCompressed.c_str());

		double hd;		// hausdorff distance
		double hdb;		// hausdorff distance with bspline
		double hd4;		// hausdorff distance with 4-point
		double hdj;		// hausdorff distance with jarek

		nSubdivision = HIGHEST_PRECISION - nXBits;
		// nSubdivision = 1;

		// bSpline
		ctemp = simplifiedNew;
		for(i = 0; i < nSubdivision; i++)
		{
			ctemp1 = ctemp.subdivided(BSPLINE);
			ctemp = ctemp1;
		}
		bSpline = ctemp;
		bSpline.writeFile(bSplineFilename.c_str());
		bSpline.draw(bSplineFilenameDraw.c_str());

		// 4 point
		ctemp = simplifiedNew;
		for(i = 0; i < nSubdivision; i++)
		{
			ctemp1 = ctemp.subdivided(FOURPOINT_SUBDIVISION);
			ctemp = ctemp1;
		}
		fourPoint = ctemp;
		fourPoint.writeFile(fourPointFilename.c_str());
		fourPoint.draw(fourPointFilenameDraw.c_str());

		// jarek
		ctemp = simplifiedNew;
		for(i = 0; i < nSubdivision; i++)
		{
			ctemp1 = ctemp.subdivided(JAREK_SUBDIVISION);
			ctemp = ctemp1;
		}
		jarek = ctemp;
		jarek.writeFile(jarekFilename.c_str());
		jarek.draw(jarekFilenameDraw.c_str());

		// hausdorff distance between the original curve and the b spline curve
		hd = hausdorff(curve, simplifiedNew);
		hdb = hausdorff(curve, bSpline);
		hd4 = hausdorff(curve, fourPoint);
		hdj = hausdorff(curve, jarek);

		cout << "Hausdorff (orig, simp) : " << hd << endl;
		cout << "Hausdorff (orig, bSpline) : " << hdb << endl;
		cout << "Hausdorff (orig, 4-Point) : " << hd4 << endl;
		cout << "Hausdorff (orig, Jarek) : " << hdj << endl;
		cout << "\n=============================================\n";
	}
	script.close();
}